@@ -19,6 +19,8 @@
 # OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 # WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 
+from __future__ import print_function
+
 import collections
 import renpy.sl2
 import renpy.sl2.slast as slast
@@ -116,7 +118,7 @@ class Parser(object):
     # inside something that takes a single child.
     nchildren = "many"
 
-    def __init__(self, name):
+    def __init__(self, name, statement=True):
 
         # The name of this object.
         self.name = name
@@ -127,7 +129,8 @@ class Parser(object):
         self.keyword = { }
         self.children = { }
 
-        all_statements.append(self)
+        if statement:
+            all_statements.append(self)
 
         global parser
         parser = self
@@ -217,6 +220,7 @@ class Parser(object):
         """
 
         seen_keywords = set()
+        block = False
 
         # Parses a keyword argument from the lexer.
         def parse_keyword(l, expect):
@@ -241,6 +245,14 @@ class Parser(object):
 
             seen_keywords.add(name)
 
+            if name == "at" and block and l.keyword("transform"):
+                l.require(":")
+                l.expect_eol()
+                l.expect_block("ATL block")
+                expr = renpy.atl.parse_atl(l.subblock_lexer())
+                target.atl_transform = expr
+                return
+
             expr = l.comma_expression()
 
             target.keyword.append((name, expr))
@@ -716,6 +728,11 @@ class ForParser(Parser):
         else:
             code = None
 
+        if l.match('index'):
+            index_expression = l.require(l.say_expression)
+        else:
+            index_expression = None
+
         l.require('in')
 
         expression = l.require(l.python_expression)
@@ -723,7 +740,7 @@ class ForParser(Parser):
         l.require(':')
         l.expect_eol()
 
-        rv = slast.SLFor(loc, name, expression)
+        rv = slast.SLFor(loc, name, expression, index_expression)
 
         if code:
             rv.children.append(slast.SLPython(loc, code))
@@ -937,7 +954,7 @@ class CustomParser(Parser):
 class ScreenParser(Parser):
 
     def __init__(self):
-        super(ScreenParser, self).__init__("screen")
+        super(ScreenParser, self).__init__("screen", statement=False)
 
     def parse(self, loc, l, parent, name="_name"):
 
