@@ -19,6 +19,8 @@
 # OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 # WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 
+from __future__ import print_function
+
 import math
 import renpy.display
 
@@ -248,6 +250,8 @@ class TextSegment(object):
 
         if self.ruby_bottom:
             textsupport.mark_ruby_bottom(rv)
+        elif self.ruby_top == "alt":
+            textsupport.mark_altruby_top(rv)
         elif self.ruby_top:
             textsupport.mark_ruby_top(rv)
 
@@ -387,7 +391,7 @@ class DisplayableSegment(object):
         w = layout.scale_int(self.width)
         h = layout.scale_int(self.height)
 
-        glyph.character = 0
+        glyph.character = 0xfffc
         glyph.ascent = 0
         glyph.line_spacing = h
         glyph.advance = w
@@ -403,14 +407,7 @@ class DisplayableSegment(object):
 
         if di.displayable_blits is not None:
 
-            xo, yo = renpy.display.core.place(
-                glyph.width,
-                glyph.ascent,
-                glyph.width,
-                glyph.line_spacing,
-                self.d.get_placement())
-
-            di.displayable_blits.append((self.d, glyph.x + xo, glyph.y + yo, glyph.time))
+            di.displayable_blits.append((self.d, glyph.x, glyph.y, glyph.width, glyph.ascent, glyph.line_spacing, glyph.time))
 
     def assign_times(self, gt, glyphs):
         if self.cps != 0:
@@ -697,11 +694,21 @@ class Layout(object):
 
         textsupport.align_and_justify(lines, maxx, style.text_align, style.justify)
 
-        if splits_from and text.style.adjust_spacing:
+        adjust_spacing = text.style.adjust_spacing
+
+        if splits_from and adjust_spacing:
             target_x = self.scale_int(splits_from.size[0] - splits_from.xborder)
             target_y = self.scale_int(splits_from.size[1] - splits_from.yborder)
 
-            textsupport.tweak_glyph_spacing(all_glyphs, lines, target_x - maxx, target_y - y, maxx, y)  # @UndefinedVariable
+            target_x_delta = target_x - maxx
+            target_y_delta = target_y - y
+
+            if adjust_spacing == "horizontal":
+                target_y_delta = 0.0
+            elif adjust_spacing == "vertical":
+                target_x_delta = 0.0
+
+            textsupport.tweak_glyph_spacing(all_glyphs, lines, target_x_delta, target_y_delta, maxx, y)  # @UndefinedVariable
 
             maxx = target_x
             y = target_y
@@ -718,7 +725,7 @@ class Layout(object):
 
         # Place ruby.
         if self.has_ruby:
-            textsupport.place_ruby(all_glyphs, self.scale_int(style.ruby_style.yoffset), sw, sh)
+            textsupport.place_ruby(all_glyphs, self.scale_int(style.ruby_style.yoffset), self.scale_int(style.altruby_style.yoffset), sw, sh)
 
         # Check for glyphs that are being drawn out of bounds, because the font
         # or anti-aliasing or whatever makes them bigger than the bounding box. If
@@ -868,203 +875,250 @@ class Layout(object):
 
         for type, text in tokens:  # @ReservedAssignment
 
-            if type == PARAGRAPH:
+            try:
 
-                # Note that this code is duplicated for the p tag, and for
-                # the empty line case, below.
-                fill_empty_line()
+                if type == PARAGRAPH:
 
-                paragraphs.append(line)
-                line = [ ]
+                    # Note that this code is duplicated for the p tag, and for
+                    # the empty line case, below.
+                    fill_empty_line()
 
-                continue
+                    paragraphs.append(line)
+                    line = [ ]
 
-            elif type == TEXT:
-                line.extend(tss[-1].subsegment(text))
-                continue
+                    continue
 
-            elif type == DISPLAYABLE:
-                line.append((DisplayableSegment(tss[-1], text, renders), u""))
-                continue
+                elif type == TEXT:
+                    line.extend(tss[-1].subsegment(text))
+                    continue
 
-            # Otherwise, we have a text tag.
+                elif type == DISPLAYABLE:
+                    line.append((DisplayableSegment(tss[-1], text, renders), u""))
+                    continue
 
-            tag, _, value = text.partition("=")
+                # Otherwise, we have a text tag.
 
-            if tag and tag[0] == "/":
-                tss.pop()
+                tag, _, value = text.partition("=")
 
-                if not tss:
-                    raise Exception("%r closes a text tag that isn't open." % text)
+                if tag and tag[0] == "/":
+                    tss.pop()
 
-            elif tag == "_start":
-                fs = FlagSegment()
-                line.append((fs, ""))
-                self.start_segment = fs
+                    if not tss:
+                        raise Exception("%r closes a text tag that isn't open." % text)
 
-            elif tag == "_end":
-                fs = FlagSegment()
-                line.append((fs, ""))
-                self.end_segment = fs
+                elif tag == "_start":
+                    fs = FlagSegment()
+                    line.append((fs, ""))
+                    self.start_segment = fs
 
-            elif tag == "p":
-                # Duplicated from the newline tag.
-                fill_empty_line()
+                elif tag == "_end":
+                    fs = FlagSegment()
+                    line.append((fs, ""))
+                    self.end_segment = fs
 
-                paragraphs.append(line)
-                line = [ ]
+                elif tag == "p":
+                    # Duplicated from the newline tag.
+                    fill_empty_line()
 
-            elif tag == "space":
-                width = self.scale_int(int(value))
-                line.append((SpaceSegment(tss[-1], width=width), u""))
+                    paragraphs.append(line)
+                    line = [ ]
 
-            elif tag == "vspace":
-                # Duplicates from the newline tag.
+                elif tag == "space":
 
-                height = self.scale_int(int(value))
+                    if len(value) < 1:
+                        raise Exception("empty value supplied for tag %r" % tag)
 
-                if line:
+                    width = self.scale_int(int(value))
+                    line.append((SpaceSegment(tss[-1], width=width), u""))
+
+                elif tag == "vspace":
+
+                    if len(value) < 1:
+                        raise Exception("empty value supplied for tag %r" % tag)
+
+                    # Duplicates from the newline tag.
+
+                    height = self.scale_int(int(value))
+
+                    if line:
+                        paragraphs.append(line)
+
+                    line = [ (SpaceSegment(tss[-1], height=height), u"") ]
                     paragraphs.append(line)
 
-                line = [ (SpaceSegment(tss[-1], height=height), u"") ]
-                paragraphs.append(line)
+                    line = [ ]
 
-                line = [ ]
+                elif tag == "w":
+                    pass
 
-            elif tag == "w":
-                pass
+                elif tag == "fast":
+                    pass
 
-            elif tag == "fast":
-                pass
+                elif tag == "nw":
+                    pass
 
-            elif tag == "nw":
-                pass
+                elif tag == "a":
+                    self.has_hyperlinks = True
 
-            elif tag == "a":
-                self.has_hyperlinks = True
+                    hyperlink_styler = style.hyperlink_functions[0]
 
-                hyperlink_styler = style.hyperlink_functions[0]
+                    if hyperlink_styler:
+                        hls = hyperlink_styler(value)
+                    else:
+                        hls = style
+
+                    old_prefix = hls.prefix
+
+                    link = len(self.hyperlink_targets) + 1
+                    self.hyperlink_targets[link] = value
+
+                    if not text_displayable.hyperlink_sensitive(value):
+                        hls.set_prefix("insensitive_")
+                    elif (renpy.display.focus.get_focused() is text_displayable) and (renpy.display.focus.argument == link):
+                        hls.set_prefix("hover_")
+                    else:
+                        hls.set_prefix("idle_")
+
+                    ts = push()
+                    # inherit vertical style
+                    vert_style = ts.vertical
+                    size = ts.size
+
+                    ts.take_style(hls, self)
+
+                    ts.vertical = vert_style
+                    ts.hyperlink = link
+
+                    if renpy.config.hyperlink_inherit_size:
+                        ts.size = size
+
+                    hls.set_prefix(old_prefix)
+
+                elif tag == "b":
+                    push().bold = True
+
+                elif tag == "i":
+                    push().italic = True
+
+                elif tag == "u":
+                    if value:
+                        push().underline = self.scale_int(int(value))
+                    else:
+                        push().underline = self.scale_int(1)
+
+                elif tag == "s":
+                    push().strikethrough = True
+
+                elif tag == "plain":
+                    ts = push()
+                    ts.bold = False
+                    ts.italic = False
+                    ts.underline = False
+                    ts.strikethrough = False
+
+                elif tag == "":
+                    style = getattr(renpy.store.style, value)
+                    push().take_style(style, self)
+
+                elif tag == "font":
+                    push().font = value
+
+                elif tag == "size":
+
+                    if len(value) < 1:
+                        raise Exception("empty value supplied for tag %r" % tag)
+
+                    if value[0] in "+-":
+                        push().size += int(value)
+                    else:
+                        push().size = int(value)
+
+                elif tag == "color":
+
+                    if len(value) < 1:
+                        raise Exception("empty value supplied for tag %r" % tag)
+
+                    push().color = renpy.easy.color(value)
+
+                elif tag == "outlinecolor":
+
+                    if len(value) < 1:
+                        raise Exception("empty value supplied for tag %r" % tag)
+
+                    push().outline_color = renpy.easy.color(value)
+
+                elif tag == "alpha":
+
+                    if len(value) < 1:
+                        raise Exception("empty value supplied for tag %r" % tag)
+
+                    ts = push()
+                    if value[0] in "+-":
+                        value = ts.color.alpha + float(value)
+                    elif value[0] == "*":
+                        value = ts.color.alpha * float(value[1:])
+                    else:
+                        value = float(value)
+
+                    ts.color = ts.color.replace_opacity(value)
+
+                elif tag == "k":
+
+                    if len(value) < 1:
+                        raise Exception("empty value supplied for tag %r" % tag)
+
+                    push().kerning = self.scale(float(value))
+
+                elif tag == "rt":
+                    ts = push()
+                    # inherit vertical style
+                    vert_style = ts.vertical
+                    ts.take_style(style.ruby_style, self)
+                    ts.vertical = vert_style
+                    ts.ruby_top = True
+                    self.has_ruby = True
+
+                elif tag == "art":
+                    ts = push()
+                    # inherit vertical style
+                    vert_style = ts.vertical
+                    ts.take_style(style.altruby_style, self)
+                    ts.vertical = vert_style
+                    ts.ruby_top = "alt"
+                    self.has_ruby = True
+
+                elif tag == "rb":
+                    push().ruby_bottom = True
+                    # We only care about ruby if we have a top.
+
+                elif tag == "cps":
+
+                    if len(value) < 1:
+                        raise Exception("empty value supplied for tag %r" % tag)
+
+                    ts = push()
+
+                    if value[0] == "*":
+                        ts.cps *= float(value[1:])
+                    else:
+                        ts.cps = float(value)
+
+                elif tag == "vert":
+                    push().vertical = True
+
+                elif tag == "horiz":
+                    ts = push()
+                    ts.vertical = False
+
+                elif tag[0] == "#":
+                    pass
 
-                if hyperlink_styler:
-                    hls = hyperlink_styler(value)
                 else:
-                    hls = style
+                    raise Exception("Unknown text tag %r" % text)
 
-                old_prefix = hls.prefix
-
-                link = len(self.hyperlink_targets) + 1
-                self.hyperlink_targets[link] = value
-
-                if not text_displayable.hyperlink_sensitive(value):
-                    hls.set_prefix("insensitive_")
-                elif (renpy.display.focus.get_focused() is text_displayable) and (renpy.display.focus.argument == link):
-                    hls.set_prefix("hover_")
-                else:
-                    hls.set_prefix("idle_")
-
-                ts = push()
-                # inherit vertical style
-                vert_style = ts.vertical
-                size = ts.size
-
-                ts.take_style(hls, self)
-
-                ts.vertical = vert_style
-                ts.hyperlink = link
-
-                if renpy.config.hyperlink_inherit_size:
-                    ts.size = size
-
-                hls.set_prefix(old_prefix)
-
-            elif tag == "b":
-                push().bold = True
-
-            elif tag == "i":
-                push().italic = True
-
-            elif tag == "u":
-                if value:
-                    push().underline = self.scale_int(int(value))
-                else:
-                    push().underline = self.scale_int(1)
-
-            elif tag == "s":
-                push().strikethrough = True
-
-            elif tag == "plain":
-                ts = push()
-                ts.bold = False
-                ts.italic = False
-                ts.underline = False
-                ts.strikethrough = False
-
-            elif tag == "":
-                style = getattr(renpy.store.style, value)
-                push().take_style(style, self)
-
-            elif tag == "font":
-                push().font = value
-
-            elif tag == "size":
-                if value[0] in "+-":
-                    push().size += int(value)
-                else:
-                    push().size = int(value)
-
-            elif tag == "color":
-                push().color = renpy.easy.color(value)
-
-            elif tag == "outlinecolor":
-                push().outline_color = renpy.easy.color(value)
-
-            elif tag == "alpha":
-                ts = push()
-                if value[0] in "+-":
-                    value = ts.color.alpha + float(value)
-                elif value[0] == "*":
-                    value = ts.color.alpha * float(value[1:])
-                else:
-                    value = float(value)
-
-                ts.color = ts.color.replace_opacity(value)
-
-            elif tag == "k":
-                push().kerning = self.scale(float(value))
-
-            elif tag == "rt":
-                ts = push()
-                # inherit vertical style
-                vert_style = ts.vertical
-                ts.take_style(style.ruby_style, self)
-                ts.vertical = vert_style
-                ts.ruby_top = True
-                self.has_ruby = True
-
-            elif tag == "rb":
-                push().ruby_bottom = True
-                # We only care about ruby if we have a top.
-
-            elif tag == "cps":
-                ts = push()
-
-                if value[0] == "*":
-                    ts.cps *= float(value[1:])
-                else:
-                    ts.cps = float(value)
-
-            elif tag == "vert":
-                push().vertical = True
-
-            elif tag == "horiz":
-                ts = push()
-                ts.vertical = False
-
-            elif tag[0] == "#":
-                pass
-
-            else:
-                raise Exception("Unknown text tag %r" % text)
+            except:
+                renpy.game.exception_info = "While processing text tag {{{!s}}} in {!r}.:".format(text, text_displayable.get_all_text())
+                raise
 
         # If the line is empty, fill it with a space.
         fill_empty_line()
@@ -1196,6 +1250,9 @@ class Layout(object):
 
         for g in l.glyphs:
 
+            if g.time == -1:
+                continue
+
             if g.time > st:
                 continue
 
@@ -1394,7 +1451,14 @@ class Text(renpy.display.core.Displayable):
 
         self._duplicatable = self.slow
 
+        # The list of displayables and their offsets.
+        self.displayable_offsets = [ ]
+
     def _duplicate(self, args):
+
+        if args and args.args:
+            args.extraneous()
+
         if self._duplicatable:
             rv = self._copy(args)
             rv._unique()
@@ -1431,6 +1495,18 @@ class Text(renpy.display.core.Displayable):
         s = s.replace("\\", "\\\\").replace("\"", "\\\"").replace("\n", "\\n")
         return u"Text \"{}\"".format(s)
 
+    def get_all_text(self):
+        """
+        Gets all the text,
+        """
+        s = u""
+
+        for i in self.text:
+            if isinstance(i, basestring):
+                s += i
+
+        return s
+
     def _scope(self, scope, update=True):
         """
         Called to update the scope, when necessary.
@@ -1534,7 +1610,10 @@ class Text(renpy.display.core.Displayable):
             text_split.append(mid_string)
 
             if self.ctc is not None:
-                text_split.append(self.ctc)
+                if isinstance(self.ctc, list):
+                    text_split.extend(self.ctc)
+                else:
+                    text_split.append(self.ctc)
 
             if end_string:
                 text_split.append(end_string)
@@ -1562,6 +1641,13 @@ class Text(renpy.display.core.Displayable):
         # Text.
         self.tokens, self.displayables = self.get_displayables(tokens)
 
+        for type, text in self.tokens:
+            if type == TAG and text.startswith("a="):
+                self.focusable = True
+                break
+        else:
+            self.focusable = False
+
     def visit(self):
 
         if self.dirty or self.displayables is None:
@@ -1636,6 +1722,12 @@ class Text(renpy.display.core.Displayable):
 
         return rv
 
+    def set_style_prefix(self, prefix, root):
+        if prefix != self.style.prefix:
+            self.kill_layout()
+
+        super(Text, self).set_style_prefix(prefix, root)
+
     def focus(self, default=False):
         """
         Called when a hyperlink gains focus.
@@ -1655,15 +1747,9 @@ class Text(renpy.display.core.Displayable):
         hyperlink_focus = self.style.hyperlink_functions[2]
         target = layout.hyperlink_targets.get(renpy.display.focus.argument, None)
 
-        if hyperlink_focus:
+        if hyperlink_focus and (not default) and (target is not None):
             return hyperlink_focus(target)
 
-    def set_style_prefix(self, prefix, root):
-        if prefix != self.style.prefix:
-            self.kill_layout()
-
-        super(Text, self).set_style_prefix(prefix, root)
-
     def unfocus(self, default=False):
         """
         Called when a hyperlink loses focus, or isn't focused to begin with.
@@ -1674,7 +1760,7 @@ class Text(renpy.display.core.Displayable):
 
         hyperlink_focus = self.style.hyperlink_functions[2]
 
-        if hyperlink_focus:
+        if hyperlink_focus and not default:
             return hyperlink_focus(None)
 
     def call_slow_done(self, st):
@@ -1715,11 +1801,17 @@ class Text(renpy.display.core.Displayable):
             raise renpy.display.core.IgnoreEvent()
 
         layout = self.get_layout()
+
         if layout is None:
             return
 
-        for d, xo, yo, _ in layout.displayable_blits:
-            rv = d.event(ev, x - xo - layout.xoffset, y - yo - layout.yoffset, st)
+        if layout.redraw_typewriter(st) is None:
+            if self.slow:
+                self.call_slow_done(st)
+                self.slow = False
+
+        for d, xo, yo in self.displayable_offsets:
+            rv = d.event(ev, x - xo, y - yo, st)
             if rv is not None:
                 return rv
 
@@ -1909,16 +2001,29 @@ class Text(renpy.display.core.Displayable):
         # Blit displayables.
         if layout.displayable_blits:
 
+            self.displayable_offsets = [ ]
+
             drend = renpy.display.render.Render(w, h)
             drend.forward = layout.reverse
             drend.reverse = layout.forward
 
-            for d, xo, yo, t in layout.displayable_blits:
+            for d, x, y, width, ascent, line_spacing, t in layout.displayable_blits:
 
                 if self.slow and t > st:
                     continue
 
-                drend.absolute_blit(renders[d], (xo + layout.xoffset, yo + layout.yoffset))
+                xo, yo = renpy.display.core.place(
+                    width,
+                    ascent,
+                    width,
+                    line_spacing,
+                    d.get_placement())
+
+                xo = x + xo + layout.xoffset
+                yo = y + yo + layout.yoffset
+
+                drend.absolute_blit(renders[d], (xo, yo))
+                self.displayable_offsets.append((d, xo, yo))
 
             rv.blit(drend, (0, 0))
 
@@ -1935,7 +2040,7 @@ class Text(renpy.display.core.Displayable):
             if redraw is not None:
                 renpy.display.render.redraw(self, redraw)
             else:
-                self.call_slow_done(st)
+                renpy.display.interface.timeout(0)
 
         rv.forward = layout.forward
         rv.reverse = layout.reverse
